#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Creation date: 2015-04-09
"""
Calculculates a list of astronomical data.

Works for a given location indicator
for a given time span. Sunrise, sunset etc. are calculated for the *night*
of interrest. so Sunset is for today, sunrise for tommorow morning etc..

..  note::

    Make sure the SQLite-database with location data can be found at
    the directory "~/.adnitevision/stations.sqlite"

Type :code:`adnite_basics.py --help` for further information."""

import sys
import argparse
import datetime

import adnitevision_iolib as aiol
import adnitevision_errlib as aerr
import adnitevision_datlib as adat
import adnitevision_constlib as aconst

__version__ = aconst.VERSION
__copyright__ = aconst.COPYRIGHT
__license__ = aconst.LICENSE

#===================== Check Args'n'Options ==========================
VERSION = __version__ + " - " + __copyright__ + \
    " - License: " + __license__  # : Program version
DESCRIPTION = u"Calculate astronomical basis data for one station and subsequent dates"
STATIONHELP = u'enter a valid location indicator, e.g. --station = LOWL. Defaults to ' + \
    aiol.LOCINDICATORDEFAULT
STATIONDEFAULT = aiol.LOCINDICATORDEFAULT
STARTDATEHELP = u'start with date yyyymmdd, e.g. startdate = 20150401 (April 1st 2015), default is today'
STARTDATEDEFAULT = adat.pydate2mydate(datetime.date.today())
ENDDATEHELP = u'end with date yyyymmdd, e.g. enddate = 20150410 (April 10th 2015), default is startdate'
ENDDATEDEFAULT = STARTDATEDEFAULT
DATESTEPHELP = u'increment of desired output, eg. datestep=2 outputs every second day, defaults to 1'
DATESTEPDEFAULT = 1  # days
VERSIONHELP = u'show program version number and copyright notice and exit'
CSVHELP = u""" output all data in CSV-format with a semicolon as data separator, e.g.,
>> 2014-05-12; 03:30; ...<<. together with a roughly self explaining header line. Have a try!"""
FILEHELP = u"""Inputfile with all relevant input data. If inputfile is given all other input parameters
except --csv are irrelevant. The file must be a csv-file with special keywords.
You should have got a copy of an sample file to see the format."""


def adnitebasics():
    """ Main routine of the program"""

    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--version', action='version',
                        help=VERSIONHELP, version='%(prog)s ' + VERSION)
    parser.add_argument(
        '-s', '--station', help=STATIONHELP, default=STATIONDEFAULT)
    parser.add_argument(
        '-a', '--startdate', help=STARTDATEHELP, default=STARTDATEDEFAULT, type=int)
    parser.add_argument(
        '-e', '--enddate', help=ENDDATEHELP, default=None, type=int)
    parser.add_argument(
        '-d', '--datestep', help=DATESTEPHELP, default=DATESTEPDEFAULT, type=int)
    parser.add_argument('--csv', action='store_true', help=CSVHELP)
    parser.add_argument('-f', '--inputfile', help=FILEHELP)
    args = parser.parse_args()

    if args.inputfile is not None:
        # U S E   I N P U T F I L E
        inputdict = aiol.readinputdatalines(args.inputfile)
        station = inputdict['STATION'].upper()
        pystartdate = inputdict['DATE1']
        try:
            pyenddate = inputdict['DATE2']
        except KeyError:
            pyenddate = pystartdate
        try:
            datestep = inputdict['DATESTEP']
        except KeyError:
            datestep = 1
    else:
        # U S E   O P T I O N S   a s   I N P U T
        if args.enddate is None:
            args.enddate = args.startdate
        else:
            if args.startdate > args.enddate:
                print("ERROR: startdate " + str(args.startdate) +
                      " greater than enddate " + str(args.enddate))
                sys.exit(aerr.ERRORS['DATEERROR'])
        station = args.station.upper()
        pystartdate = adat.mydate2pydate(args.startdate)
        pyenddate = adat.mydate2pydate(args.enddate)
        datestep = args.datestep

    statdat = aiol.readstationlistdb(station)
    #=============== Standard-Output headers and formats ================
    header0 = u"Station: {}  {}".format(statdat['station'], statdat['name'])
    header1 = u"Latitude: {:8.4f}  Longitude: {:7.4f} degree   Elevation: {:4.0f} m".format(
        statdat['lat'], statdat['lon'], statdat['elev'])
    header2 = u"Start-Date: {}   End-Date {}    All times in UTC".format(
        pystartdate, pyenddate)
    caption = ("      DATE |      SSET      ECET      ENET      BNMT      BCMT     SRISE " +
               "|     Moon  (rise, set)  Phase  Next Full  New Moon")
    doubleline = 124 * '='

    #================ CSV-output headers and formats ==================
    csv_header = ("locind; date; ssdate; sstime; ecetdate; ecettime; enetdate; enettime; " +
                  "bmntdate; bmnttime; bcmtdate; bcmttime; srdate; srtime; " +
                  "srflag1; moon1date; moon1time; srflag2; moon2date; moon2time; " +
                  "moonphase; nextfullm; nextnewm")

    day = pystartdate
    if not args.csv:
    #================ Output-Standard ===============================
        print(doubleline)
        print header0
        print(header1)
        print(header2)
        print(doubleline)
        print(caption)
        while day <= pyenddate:
            aiol.astro_dayline(station, day)
            day = day + datetime.timedelta(days=datestep)
        print(doubleline)
    else:
    #================ Output-CSV ===============================
        print(csv_header)
        while day <= pyenddate:
            aiol.astro_dayline_csv(station, day)
            day = day + datetime.timedelta(days=datestep)

if __name__ == "__main__":
    adnitebasics()
