#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Creation date: 2015-04-28
""" Outputs a sheet with both text and graphic information for basic
astronomic, moon position and night illuminances dependimng on
location, time and environment.

..  note::

    Make sure the SQLite-database with location data can be found at
    the directory "~/.adnitevision/stations.sqlite"

Type :code:`adnite_basics.py --help` for further information."""


import argparse
import datetime
import numpy as np
import matplotlib.pyplot as plt
import os.path
import sys

import adnitevisionlib as al
import adnitevision_iolib as aiol
# import adnitevision_errlib as aerr
import adnitevision_datlib as adat
import adnitevision_constlib as aconst

__version__ = aconst.VERSION
__copyright__ = aconst.VERSION
__license__ = aconst.LICENSE

HEADERFONTSIZE = 14
FOOTERFONTSIZE = 10
TEXTFONTSIZE = 11
FINETIMESTEP = 5  # : minutes (for graphic output)

#=========================== Args'n'Options ==========================
VERSION = __version__ + " - " + __copyright__ + \
    " - License: " + __license__  # : Program version
DESCRIPTION = u"Calculculates basic astronomic data, moon data and night illuminances dependent on time and location."
STATIONHELP = u'End a valid location indicator, e.g. station = LOWL'
STATIONDEFAULT = aiol.LOCINDICATORDEFAULT
STARTDATEHELP = u'Start with date yyyymmdd, e.g. startdate = 20150401 (April 1st 2015), default is today'
STARTDATEDEFAULT = adat.pydate2mydate(datetime.date.today())
STARTTIMEHELP = u'starts with ddhh in UTC,  e.g. --starttime = 1800, default is 1700'
STARTTIMEDEFAULT = 1700
ENDTIMEHELP = u'ends with <= ddhh in UTC (depending an timestep),  e.g. --endtime = 0600, default is 0500'
ENDTIMEDEFAULT = int('0500')
TIMESTEPHELP = u'Increment of desired output in minutes, e.g. --timestep=30 outputs every half hour, defaults to 120.'
TIMESTEPDEFAULT = 120  # minutes
VERSIONHELP = u'Shows programs version number and copyright notice and exit'
CLOUDINESSHELP = """input the simplified overall cloudiness as a constant in time in eights of the complete sky;
e.g. cloudiness=0 means no sky clear, cloudiness=8 means overcast."""
CLOUDINESSDEFAULT = 0  # no clouds as default
FILEHELP = u"""Inputfile with all relevant input data. If inputfile is given all other input parameters
except --csv are irrelevant. The file must be a csv-file with special keywords.
You should have got a copy of an sample file to see the format."""
GRAPHOUTDIR = './'
GRAPHOUTDIRHELP = u"""Output-Directory for the resulting graphic file.
Defaults to actual directory."""
GRAPHTYPES = ['pdf', 'png', 'jpg']
GRAPHTYPEDEFAULT = 'pdf'
GRAPHTYPEHELP = u"Graphic type of the resulting output graphic. Defaults to "\
                + str(GRAPHTYPEDEFAULT)


def read_options_arguments():
    """Defines and reads options and arguments"""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--version', action='store_true',
                        help=VERSIONHELP)  # ,version='%(prog)s ' + VERSION)
    parser.add_argument(
        '-s',
        '--station',
        help=STATIONHELP,
        default=STATIONDEFAULT)
    parser.add_argument(
        '-a',
        '--startdate',
        help=STARTDATEHELP,
        default=STARTDATEDEFAULT,
        type=int)
    parser.add_argument(
        '-t',
        '--starttime',
        help=STARTTIMEHELP,
        default=STARTTIMEDEFAULT,
        type=int)
    parser.add_argument(
        '-T',
        '--endtime',
        help=ENDTIMEHELP,
        default=ENDTIMEDEFAULT,
        type=int)
    parser.add_argument(
        '-d',
        '--timestep',
        help=TIMESTEPHELP,
        default=TIMESTEPDEFAULT,
        type=float)
    parser.add_argument(
        '-N',
        '--cloudiness',
        help=CLOUDINESSHELP,
        default=CLOUDINESSDEFAULT,
        type=int)
    parser.add_argument('-f', '--inputfile', help=FILEHELP)
    parser.add_argument(
        '-o',
        '--graphoutdir',
        help=GRAPHOUTDIRHELP,
        default=GRAPHOUTDIR)
    parser.add_argument(
        '-g',
        '--graphtype',
        help=GRAPHTYPEHELP,
        default=GRAPHTYPEDEFAULT,
        choices=GRAPHTYPES)

    args = parser.parse_args()
    return args


def use_options(args):
    """Process input options"""
    station = args.station.upper()
    pystartdate = adat.mydate2pydate(args.startdate)
    starttime = int(args.starttime)
    endtime = int(args.endtime)
    timestep = float(args.timestep)
    pystartdatetime = adat.combinepydatepytime(
        pystartdate, adat.hhmm2pytime(starttime))
    if endtime > starttime:
        pyenddatetime = adat.combinepydatepytime(
            pystartdate, adat.hhmm2pytime(endtime))
    else:
        pyenddatetime = (adat.combinepydatepytime(pystartdate,
                         adat.hhmm2pytime(endtime)) +
                         adat.pytimedelta(dd=1))
    pytimestep = adat.pytimedelta(mi=timestep)
    pydatetime_array = aiol.makepydatetimelist(
        pystartdatetime,
        pyenddatetime,
        pytimestep)
    cloudiness = max(min(args.cloudiness, 10), 0)  # limit allowed cloudiness
    environment = (cloudiness * np.ones(len(pydatetime_array)),)
    return (station, pystartdate, pystartdatetime, pyenddatetime,
            pytimestep, pydatetime_array, environment)


def use_inputfile(inputfile):
    """Process input file"""
    inputdict = aiol.readinputdatalines(inputfile)
    return aiol.process_inputdict(inputdict)


def make_environment_fine(pydatetime_fine, pydatetime, environment):
    clouds = environment[0]
    ephemdate = adat.pydatetime2ephemdate_array(pydatetime)
    ephemdate_fine = adat.pydatetime2ephemdate_array(pydatetime_fine)
    clouds_fine = np.interp(ephemdate_fine, ephemdate, clouds)
    return (clouds_fine,)


def adnite_graph():
    """ Main routine of the program"""
    args = read_options_arguments()
    if args.version:  # if "--version" print version and exit, because built
        aconst.version('Program')  # in functionality of argparse is ugly!
        sys.exit(0)
    graphoutdir = args.graphoutdir
    if not os.path.exists(graphoutdir):
        print("Warning: output directory for graphic result")
        print(graphoutdir)
        print("does not exist.")
        print("Set to default: " + GRAPHOUTDIR)
        graphoutdir = GRAPHOUTDIR

    # Input
    if args.inputfile:  # input by file
        (station, pystartdate, pystartdatetime, pyenddatetime,
         pytimestep, pydatetime, environment) = use_inputfile(args.inputfile)
    else:  # input by options
        (station, pystartdate, pystartdatetime, pyenddatetime,
         pytimestep, pydatetime, environment) = use_options(args)
    statdat = aiol.readstationlistdb(station)

    # calculate   c o a r s e   data for text
    datetimeutc = adat.Date(pystartdate)  # convert in pypephem date format
    locnoon, locmidnight, srs, ctw, ntw, atw, mrs, mphase, mfn, nfm, nnm =\
        al.calc_astro_night(statdat, datetimeutc)
    moondata = al.moondata(statdat, pydatetime)
    sundata = al.sundata(statdat, pydatetime)
    mdates, maltitude, mazimuth, msize, mmag, mphase = moondata
    #
    # print mmag
    # print al.mag2lux(mmag)
    #
    sdates, saltitude, sazimuth, ssize, smag = sundata
    (m_illum, s_illum, twl_illum, brutto_illum, netto_illum) =\
        al.calc_illumination(moondata, sundata, environment)

    # calculate   f i n e   data for graphic
    pydatetimef = adat.pydatetime_array(pystartdatetime,
                                        pyenddatetime, d_minutes=FINETIMESTEP)
    environmentf = make_environment_fine(pydatetimef,
                                         pydatetime, environment)
    cloudsf = environmentf[0]
    moondataf = al.moondata(statdat, pydatetimef)
    sundataf = al.sundata(statdat, pydatetimef)
    mdatesf, maltitudef, mazimuthf, msizef, mmagf, mphasef = moondataf
    sdatesf, saltitudef, sazimuthf, ssizef, smagf = sundataf
    # real environment (clouds, ...)
    (m_illumf, s_illumf, twl_illumf, brutto_illumf, netto_illumf) =\
        al.calc_illumination(moondataf, sundataf, environmentf)
    # clouds N=0
    cloudsf0 = np.ones(len(cloudsf)) * 0.0
    environmentf0 = (cloudsf0,)
    (m_illumf, s_illumf, twl_illumf, brutto_illumf, netto_illumf0) =\
        al.calc_illumination(moondataf, sundataf, environmentf0)
    # clouds N=8
    cloudsf8 = np.ones(len(cloudsf)) * 8.0
    environmentf8 = (cloudsf8,)
    (m_illumf, s_illumf, twl_illumf, brutto_illumf, netto_illumf8) =\
        al.calc_illumination(moondataf, sundataf, environmentf8)

    # Prepare output strings
    header0 = u"Station: {} {} Date: {}".format(
        statdat['station'], 37 * ' ', pystartdate)
    header01 = u"{}".format(statdat['name'])
    header1 = u"Latitude: {:8.4f}  Longitude: {:7.4f} degree   Elevation: {:4.0f} m".format(
        statdat['lat'], statdat['lon'], statdat['elev'])
    header2 = u"Start-Time: {}   End-Time: {}  (UTC)".format(
        pystartdatetime.strftime("%H:%M"), pyenddatetime.strftime("%H:%M"))
    caption_astro1 = "    SSET      ECET      ENET      BNMT      BCMT     SRISE "
    caption_astro2 = " Moon  (rise, set)      Phase  Next Full  New Moon"
    astro_line1, astro_line2 = aiol.astro_dayline(
        station, pystartdate, printout=False)
    outplines_moon = aiol.moon_daylines(moondata)
    outplines_ill = aiol.illum_daylines(moondata, sundata, environment)
    separator = len(outplines_moon[0]) * '-'

    # Text output to console (stdio)
    print separator
    print header0
    print header01
    print header1
    print header2
    print separator
    print caption_astro1
    print astro_line1
    print separator
    print caption_astro2
    print astro_line2
    print separator
    moon_line = ''
    for line in outplines_moon:
        moon_line = moon_line + line.strip() + '\n'
    print moon_line.strip()
    print separator
    illum_line = ''
    for line in outplines_ill:
        illum_line = illum_line + line.strip() + '\n'
    print illum_line.strip()

    # plot
    if True:
        # preamble
        plt.figure(figsize=(21. / 2.54, 29.7 / 2.54),
                   facecolor = 'mintcream')  # A4
        plt.figtext(0.5, 0.02, aconst.main_footerline(
                    os.path.basename(sys.argv[0])),
                    size=FOOTERFONTSIZE,
                    horizontalalignment='center',
                    fontweight='normal')

        #>>>>>>>>>>>>> subplot with text
        plt.subplot(2, 1, 1)
        ax1 = plt.gca()
        plt.text(0., 1.0,
                 header0 + "\n" + header01,
                 size=HEADERFONTSIZE, fontweight='bold')
        plt.text(0., 0.85,
                 header1 + "\n" + header2, size=TEXTFONTSIZE,
                 family='sans-serif',
                 fontweight='bold')
        plt.text(0., 0.65,
                 separator + '\n' + caption_astro1 + '\n' + astro_line1,
                 size=TEXTFONTSIZE, family='monospace',
                 fontweight='semibold')
        plt.text(0., 0.43,
                 separator + '\n' + caption_astro2 + '\n' + astro_line2 + '\n',
                 size=TEXTFONTSIZE,
                 family='monospace',
                 fontweight='semibold')
        plt.text(0., 0.22, moon_line.strip() + '\n' + separator,
                 size=TEXTFONTSIZE,
                 family='monospace',
                 fontweight='semibold')
        plt.text(0., -0., illum_line.strip() + '\n' + separator,
                 size=TEXTFONTSIZE,
                 family='monospace',
                 fontweight='semibold')

        ax1.axis('off')

        #>>>>>>>>>>> subplot with graphic
        plt.subplot(2, 1, 2)
        cloudsf = environmentf[0]
        starttime = np.modf(sdates[0])[0]
        pldates = ((sdatesf - sdatesf[0]) * 24.0) + starttime * 24 + 12.
        xtickdata = np.arange(pldates[0], pldates[-1] + 1)
        xticklabel = []
        for x in xtickdata:
            x1 = x % 24
            xticklabel.append("{:02.0f}".format(x1))

        illumlabels = [0.1, 1, 10, 100, 1000, 1e4]
        plt.semilogy(pldates, netto_illumf, lw=2, c='g')
        plt.semilogy(pldates, netto_illumf0, lw=1, c='k', ls=':')
        plt.semilogy(pldates, netto_illumf8, lw=1, c='k', ls=':')
        plt.semilogy(pldates,
                     np.ones(len(pldates))*aconst.MILILUXALERT_LOWER,
                     c='r', ls=':', lw=2)
        plt.ylim(1e-1, 1e4)
        plt.xlim(pldates[0], pldates[-1:])
        plt.xticks(xtickdata, xticklabel)
        plt.yticks(illumlabels, illumlabels)
        plt.xlim(pldates[0], pldates[-1])
        plt.xlabel('Time [UTC]')
        plt.ylabel('Illuminance [mlx]')
        plt.grid()
        plt.minorticks_on()
        goutfname = os.path.join(
            graphoutdir,
            statdat['station'] + '_' +\
            str(pystartdate) + "." + args.graphtype)
        plt.savefig(goutfname, facecolor='mintcream')
        print separator
        print("Graphic output saved to " + goutfname)


if __name__ == "__main__":
    adnite_graph()
