#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Creation date: 2015-04-16
"""
Calculculates night illumination by twilight, moon and dark sky luminances
dependent on time and location. Input will be delivered by command line
options, for time dependent input like cloudiness also as a file to be read.

..  note::

    Make sure the SQLite-database with location data can be found at
    the directory "~/.adnitevision/stations.sqlite"

Type :code:`adnite_illuminance.py --help` for further information."""

import argparse
import datetime
import numpy as np

import adnitevisionlib as al
import adnitevision_iolib as aiol
import adnitevision_errlib as aerr
import adnitevision_datlib as adat

__version__ = aerr.VERSION
__copyright__ = aerr.COPYRIGHT
__license__ = aerr.LICENSE

#===================== Check Args'n'Options ==========================
VERSION = __version__ + " - " + __copyright__ + \
    "- License: " + __license__  # : Program version
DESCRIPTION = u"Calculculates night illumination by twilight, moon and dark sky luminances dependent on time and location."
STATIONHELP = u'End a valid location indicator, e.g. station = LOWL'
STATIONDEFAULT = aiol.LOCINDICATORDEFAULT
STARTDATEHELP = u'Start with date yyyymmdd, e.g. startdate = 20150401 (April 1st 2015), default is today'
STARTDATEDEFAULT = adat.pydate2mydate(datetime.date.today())
STARTTIMEHELP = u'starts with ddhh in UTC,  e.g. --starttime = 1800, default is 1700'
STARTTIMEDEFAULT = 1700
ENDTIMEHELP = u'ends with <= ddhh in UTC (depending an timestep),  e.g. --endtime = 0600, default is 0500'
ENDTIMEDEFAULT = int('0500')
TIMESTEPHELP = u'Increment of desired output in minutes, e.g. --timestep=30 outputs every half hour, defaults to 120.'
TIMESTEPDEFAULT = 120  # minutes
VERSIONHELP = u'Shows programs version number and copyright notice and exit'
CSVHELP = u""" Output all data in the CSV-format with a semicolon as data separator, e.g.,
>> 2014-05-12; 03:30; ...<<. together with an roughly self explaining header line. Have a try!"""
CLOUDINESSHELP = """input the simplified overall cloudiness as a constant in time in eights of the complete sky;
e.g. cloudiness=0 means no sky clear, cloudiness=8 means overcast."""
CLOUDINESSDEFAULT = 0  # no clouds as default
FILEHELP = u"""Inputfile with all relevant input data. If inputfile is given all other input parameters
except --csv are irrelevant. The file must be a csv-file with special keywords.
You should have got a copy of an sample file to see the format."""


def use_inputfile(inputfile):
    """Process input file"""
    inputdict = aiol.readinputdatalines(inputfile)
    return aiol.process_inputdict(inputdict)


def adnite_illuminance():
    """ Main routine of the program"""

    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--version', action='version',
                        help=VERSIONHELP, version='%(prog)s ' + VERSION)
    parser.add_argument(
        '-s',
        '--station',
        help=STATIONHELP,
        default=STATIONDEFAULT)
    parser.add_argument(
        '-a',
        '--startdate',
        help=STARTDATEHELP,
        default=STARTDATEDEFAULT,
        type=int)
    parser.add_argument(
        '-t',
        '--starttime',
        help=STARTTIMEHELP,
        default=STARTTIMEDEFAULT,
        type=int)
    parser.add_argument(
        '-T',
        '--endtime',
        help=ENDTIMEHELP,
        default=ENDTIMEDEFAULT,
        type=int)
    parser.add_argument(
        '-d',
        '--timestep',
        help=TIMESTEPHELP,
        default=TIMESTEPDEFAULT,
        type=float)
    parser.add_argument(
        '-N',
        '--cloudiness',
        help=CLOUDINESSHELP,
        default=CLOUDINESSDEFAULT,
        type=int)
    parser.add_argument('--csv', action='store_true', help=CSVHELP)
    parser.add_argument('-f', '--inputfile', help=FILEHELP)
    args = parser.parse_args()

    #============= Check Args'n'Options ==============================
    if args.inputfile is None:
        # U S E   O P T I O N S
        pydate = adat.mydate2pydate(args.startdate)
        starttime = int(args.starttime)
        endtime = int(args.endtime)
        timestep = float(args.timestep)
        station = args.station.upper()
        pystartdatetime = adat.combinepydatepytime(
            pydate, adat.hhmm2pytime(starttime))
        if endtime > starttime:
            pyenddatetime = adat.combinepydatepytime(
                pydate, adat.hhmm2pytime(endtime))
        else:
            pyenddatetime = (adat.combinepydatepytime(pydate,
                             adat.hhmm2pytime(endtime)) +
                             adat.pytimedelta(dd=1))
        pytimestep = adat.pytimedelta(mi=timestep)
        cloudiness = args.cloudiness
        pydatetime = aiol.makepydatetimelist(
            pystartdatetime,
            pyenddatetime,
            pytimestep)
    else:
        # U S E   I N P U T F I L E
        (station, pystartdate, pystartdatetime, pyenddatetime,
         pytimestep, pydatetime, environment) = use_inputfile(args.inputfile)

    statdat = aiol.readstationlistdb(station)
    #=============== Standard-Output headers and formats ================
    header0 = u"Station: {}  {}".format(
        statdat['station'],
        statdat['name']) + "    --->  ILLUMINANCES"
    header1 = u"Latitude: {:8.4f}  Longitude: {:7.4f} degree   Elevation: {:4.0f} m".format(
        statdat['lat'], statdat['lon'], statdat['elev'])
    header2 = u"Start-Time: {}   End-Time {}    Timestep {}  (Times in UTC)".format(pystartdatetime,
                                                                                    pyenddatetime, pytimestep)

    #================ CSV-output headers and formars ==================
    csvheader1 = u"station; {}; name; {}".format(
        statdat['station'], statdat['name'])
    csvheader2 = u"latitude; {:8.4f}; longitude; {:8.4f}; elevation; {:5.0f}".format(
        statdat['lat'], statdat['lon'], statdat['elev'])

    moondata = al.moondata(statdat, pydatetime)
    sundata = al.sundata(statdat, pydatetime)
    mdates, maltitude, mazimuth, msize, mmag, mphase = moondata

    # NO INPUTFILE -> TIME CONSTANT ENVIRONMENT
    if args.inputfile is None:
        # actually only cloudiness is take as environment variables
        environment = np.ones(len(moondata[1]))
        environment = (cloudiness * environment,)

    if not args.csv:
        outplines = aiol.illum_daylines(moondata, sundata, environment)
        HLMax = max(
            len(outplines[0]),
            len(header0),
            len(header1),
            len(header2))
        doubleline = HLMax * '='
        print(doubleline)
        print(header0)
        print(header1)
        print(header2)
        print(doubleline)
        for line in outplines:
            print(line.strip())
        print(doubleline)
    else:
        outplines = aiol.illum_daylines_csv(moondata, sundata, environment)
        print(csvheader1)
        print(csvheader2)
        for line in outplines:
            print(line.strip())

if __name__ == "__main__":
    adnite_illuminance()
