#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Creation date: 2015-04-09
"""
Calculculates moon height, azimuth and other related data
between a start date and time to end date and an end time for a given
location indicator.

..  note::

    Make sure the SQLite-database with location data can be found at
    the directory "~/.adnitevision/stations.sqlite"

Type :code:`adnite_moon.py --help` for further information.
"""

import sys
import argparse
import datetime

import adnitevisionlib as al
import adnitevision_iolib as aiol
import adnitevision_errlib as aerr
import adnitevision_datlib as adat

__version__ = aerr.VERSION
__copyright__ = aerr.COPYRIGHT
__license__ = aerr.LICENSE

#===================== Check Args'n'Options ==========================
VERSION = __version__ + " - " + __copyright__ + \
    " - License: " + __license__  # : Program version
DESCRIPTION = u"Calculates moon postion for a location starting at a given date between two given times (UTC)"
STATIONHELP = u'enter a valid location indicator, e.g. station = LOWL. Defaults to ' + \
    aiol.LOCINDICATORDEFAULT
STATIONDEFAULT = aiol.LOCINDICATORDEFAULT
STARTDATEHELP = u'start with date yyyymmdd, e.g. startdate = 20150401 (April 1st 2015), default is today'
STARTDATEDEFAULT = adat.pydate2mydate(datetime.date.today())
STARTTIMEHELP = u'start with ddhh in UTC,  e.g. --starttime = 1800, default is 1700'
STARTTIMEDEFAULT = 1700
ENDTIMEHELP = u'end with <= ddhh in UTC (depending an timestep),  e.g. --endtime = 0600, default is 0500'
ENDTIMEDEFAULT = int('0500')
TIMESTEPHELP = u'increment of desired output in minutes, e.g. --timestep=30 outputs every half hour, defaults to 120.'
TIMESTEPDEFAULT = 120  # minutes
VERSIONHELP = u'show program version number and copyright notice and exit'
CSVHELP = u""" output all data in CSV-format with a semicolon as data separator, e.g.,
>> 2014-05-12; 03:30; ...<<. together with a roughly self explaining header line. Have a try!"""
FILEHELP = u"""Inputfile with all relevant input data. If inputfile is given all other input parameters
except --csv are irrelevant. The file must be a csv-file with special keywords.
You should have got a copy of an sample file to see the format."""


def adnitemoon():
    """ Main routine of the program"""

    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--version', action='version',
                        help=VERSIONHELP, version='%(prog)s ' + VERSION)
    parser.add_argument(
        '-s',
        '--station',
        help=STATIONHELP,
        default=STATIONDEFAULT)
    parser.add_argument(
        '-a',
        '--startdate',
        help=STARTDATEHELP,
        default=STARTDATEDEFAULT,
        type=int)
    parser.add_argument(
        '-t',
        '--starttime',
        help=STARTTIMEHELP,
        default=STARTTIMEDEFAULT,
        type=int)
    parser.add_argument(
        '-T',
        '--endtime',
        help=ENDTIMEHELP,
        default=ENDTIMEDEFAULT,
        type=int)
    parser.add_argument(
        '-d',
        '--timestep',
        help=TIMESTEPHELP,
        default=TIMESTEPDEFAULT,
        type=float)
    parser.add_argument('-f', '--inputfile', help=FILEHELP)
    parser.add_argument('--csv', action='store_true', help=CSVHELP)
    args = parser.parse_args()

    #============= Check Args'n'Options ==============================

    if args.inputfile is not None:
        # READ DATA from INPUTFILE
        inputdict = aiol.readinputdatalines(args.inputfile)
        station = inputdict['STATION'].upper()
        pydate = inputdict['DATE1']
        times = inputdict['TIMES']
        t1 = times[0]  # first time in list
        t2 = times[-1]  # last time in list
        if t2 >= t1:
            print("ERROR: " + str(t2) + " >= " + str(t1))
            print ("Time span 24 hours or more not allowed!")
            print("Check inputfile. - Exit!")
            sys.exit(aerr.ERRORS['TIMESPANERROR'])
        pydatetime = []
        pydate1 = pydate  # old day
        pydate2 = pydate + datetime.timedelta(days=1)  # new day
        for t in times:
            pydate = pydate1
            if t < t1:  # change of date
                pydate = pydate2
            ptd = adat.combinepydatepytime(pydate, t)
            pydatetime.append(ptd)
        pystartdatetime = pydatetime[0]
        pyenddatetime = pydatetime[-1]
        pytimestep = None
    else:
        # USE INPUTARGUMENTS
        pydate = adat.mydate2pydate(args.startdate)
        starttime = int(args.starttime)
        endtime = int(args.endtime)
        timestep = float(args.timestep)
        station = args.station.upper()
        pystartdatetime = adat.combinepydatepytime(
            pydate, adat.hhmm2pytime(starttime))
        if endtime > starttime:
            pyenddatetime = adat.combinepydatepytime(
                pydate, adat.hhmm2pytime(endtime))
        else:
            pyenddatetime = (adat.combinepydatepytime(pydate,
                             adat.hhmm2pytime(endtime)) +
                             adat.pytimedelta(dd=1))
        pytimestep = adat.pytimedelta(mi=timestep)
        pydatetime = aiol.makepydatetimelist(pystartdatetime, pyenddatetime,
                                             pytimestep)
    statdat = aiol.readstationlistdb(station)

    #=============== Standard-Output headers and formats ================
    header0 = u"Station: {}  {}".format(
        statdat['station'],
        statdat['name']) + "    --->  MOON-POSITION"
    header1 = u"Latitude: {:8.4f}  Longitude: {:7.4f} degree   Elevation: {:4.0f} m".format(
        statdat['lat'], statdat['lon'], statdat['elev'])
    header2 = u"Start-Time: {}   End-Time: {}   Timestep: {}  (Times in UTC)".format(pystartdatetime,
                                                                                     pyenddatetime, pytimestep)

    #================ CSV-output headers and formars ==================
    csvheader1 = u"station; {}; name; {}".format(
        statdat['station'], statdat['name'])
    csvheader2 = u"latitude; {:8.4f}; longitude; {:8.4f}; elevation; {:5.0f}".format(
        statdat['lat'], statdat['lon'], statdat['elev'])

    moondata = al.moondata(statdat, pydatetime)
    mdates, maltitude, mazimuth, msize, mmag, mphase = moondata

    if not args.csv:
        outplines = aiol.moon_daylines(moondata)
        HLMax = max(
            len(outplines[0]),
            len(header0),
            len(header1),
            len(header2))
        # print HLMax
        doubleline = HLMax * '='
        print(doubleline)
        print(header0)
        print(header1)
        print(header2)
        print(doubleline)
        for line in outplines:
            print(line.strip())
        print(doubleline)
    else:
    #================ Output-CSV ===============================
        outplines = aiol.moon_daylines_csv(moondata)
        print(csvheader1)
        print(csvheader2)
        for line in outplines:
            print(line.strip())

if __name__ == "__main__":
    adnitemoon()
