#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Creation date: 2015-04-30
"""Provides some usefull routines for date-time manipulation und
conversion for standard datetime objects and pyephem date objects."""

from datetime import date, datetime, timedelta, time
import ephem
import sys

import adnitevision_errlib as aerr
import adnitevision_constlib as aconst

def mydate2pydate(ymd):
    """Converts integers of yyyymmdd, e.g. 20150410, to
    standard python dates"""
    h = int(ymd)
    yy = h / 10000
    dd = h % 100
    mm = (h / 100) % 100
    try:
        dt = date(yy, mm, dd)
    except ValueError:
        print("ERROR: " + str(h) + " is an invalid date. Check input.")
        sys.exit(aerr.ERRORS['DATEERROR'])
    return dt


def pydate2mydate(pydat):
    """Converts standard python dates to integers yyyymmdd,
    e.g. 20150410. - Deprecated: use adnitevioson_datlib"""
    return pydat.year * 10000 + pydat.month * 100 + pydat.day


def hhmm2pytime(hhmm):
    """Converts hhmm to hours and fractions of it.
    No check for valid minutes is performed!
    - Deprecated: use adnitevioson_datlib"""
    h = int(hhmm) / 100
    m = int(hhmm) % 100
    try:
        time1 = time(hour=h, minute=m)
    except:
        print("ERROR: invalid time " + str(hhmm))
        print("Check it. - Exit")
        sys.exit(aerr.ERRORS['TIMEERROR'])
    return time1


def combinepydatepytime(pydate, pytime):
    """Wrapper for datetime.datetime.combine()"""
    return datetime.combine(pydate, pytime)


def pytimedelta(dd=0, hh=0, mi=0, ss=0.0):
    """ Wrapper for datetime.timedelta"""
    return timedelta(days=dd, hours=hh, minutes=mi, seconds=ss)


def pydatetime_array(datetime1, datetime2, d_days=0,
                     d_hours=0, d_minutes=0., d_seconds=0.):
    """Constructs an array/a list of python datetime objects with an
    increment of d_days,d_hours,d_minutes and d_seconds (all in integer,
    s_seconds in float)"""
    dati_arr = []
    dati = datetime1
    datidelta = timedelta(days=d_days, hours=d_hours,
                          minutes=d_minutes, seconds=d_seconds)
    if datidelta > timedelta(0):
        while dati <= datetime2:
            dati_arr.append(dati)
            dati = dati + datidelta
    return dati_arr


def pydatetime2ephemdate_array(pydatetime_array):
    """Converts an array/a list of python datetime objects to
    an array / a list of pyephem dates."""
    ephemdate = [ephem.Date(d) for d in pydatetime_array]
    return ephemdate


def ephemdate2pydatetime(ephemdate, roundit=3):
    """Converts ephem datetime to python datetime exact within
    roundit decimal digits of secondes, defaults to 3 meaning 0.001 s."""
    ed = ephem.Date(ephemdate)
    y, M, d, h, m, s = ed.tuple()
    pd = datetime(y, M, d, h, m) + timedelta(seconds=round(s, roundit))
    return pd


def ephemdate2pydatetime_array(ephemdatearray, roundit=3):
    """Converts ephem datetime arrays to python datetime arrays exact
    within oundit decimal digits of secondes, defaults to 3
    meaning 0.001 s."""
    pd_array = []
    for ed in ephemdatearray:
        pd = ephemdate2pydatetime(ed, roundit)
        pd_array.append(pd)
    return pd_array


def Date(datetime_str):
    """Passing thru the ephem.Date function. Creates a Date object"""
    return ephem.Date(datetime_str)


def ephemdatetime2pydatetime(ephemdatetime, roundit=3):
    """Converts ephem datetime to python datetime exact within
    roundit decimal digits of secondes, defaults to 3 meaning 0.001 s."""
    ed = ephem.Date(ephemdatetime)
    y, M, d, h, m, s = ed.tuple()
    pd = datetime(y, M, d, h, m) + timedelta(seconds=round(s, roundit))
    return pd


def ephemdatetime2pydatetime_arrays(ephemdatetimearray, roundit=3):
    """Converts ephem datetime arrays to python datetime arrays exact
    within oundit decimal digits of secondes, defaults to 3
    meaning 0.001 s."""
    pd_array = []
    for ed in ephemdatetimearray:
        pd = ephemdatetime2pydatetime(ed, roundit)
        pd_array.append(pd)
    return pd_array


if __name__ == "__main__":
    aconst.version()
    print
    yy1 = 2015
    mo1 = 4
    dd1 = 4
    hh1 = 17
    mi1 = 0
    ss1 = 0
    yy2 = 2015
    mo2 = 4
    dd2 = 5
    hh2 = 05
    mi2 = 0
    ss2 = 0
    timedelta_minutes = 120
    # Test 1
    print "Test pydatetime_array"
    pd1 = datetime(yy1, mo1, dd1, hh1, mi1, ss1)
    pd2 = datetime(yy2, mo2, dd2, hh2, mi2, ss2)
    print "From:", pd1, " Till: ", pd2
    pdt_arr = pydatetime_array(pd1, pd2, d_minutes=timedelta_minutes)
    for pd in pdt_arr:
        print "Py", pd
    print

    # Test 2
    print "Test pydatetime2ephemdate_array"
    ephem_dt_arr = pydatetime2ephemdate_array(pdt_arr)
    for ed in ephem_dt_arr:
        y, M, d, h, m, s = ed.tuple()
        print "Ephem {:4d}-{:02d}-{:02d}  {:02d}:{:02d}:{:02.0f}".\
            format(y, M, d, h, m, s)
    print

    # Test 3
    print "Test ephemdate2pydatetime_array"
    py_arr_new = ephemdate2pydatetime_array(ephem_dt_arr)
    for pd in py_arr_new:
        print "PNew", pd
