#!/usr/bin/env python2
# -*- coding: utf-8 -*-
# Creation date: 2014-11-26

"""Module contains **input-output routines** for the adnitevision program
suite.
"""

import sys
import csv
import os.path as op
import copy
import numpy as np
from datetime import date, time, datetime, timedelta


import sqlite3 as sq

import adnitevisionlib as al
import adnitevision_errlib as aerr
import adnitevision_datlib as adat
import adnitevision_constlib as aconst

__version__ = aerr.VERSION
__copyright__ = aerr.COPYRIGHT
__license__ = aerr.LICENSE

VERSION = aerr.VERSION  # :  global version
STATFILE = './stationslist.txt'  # not actively needed
STATIONDATABASE = aconst.STATIONDATABASE  # : sqlite station database
LOCINDICATORDEFAULT = 'LOWG'  # : default location indicator

ALLOWED_INPUTFILE_KEYWORDS = (
    'STATION',
    'DATE1',
    'DATE2',
    'DATESTEP',
    'TIMES',
    'CLOUDS')


def str2date(datstr):
    """ Converts the string "2014-01-31" to a python datetime.date object
    with the meaning "31 Jan 2014" """
    yy = int(datstr[0:4])
    mm = int(datstr[5:7])
    dd = int(datstr[8:10])
    return date(yy, mm, dd)


def str2hhmm(timestring):
    """ Converts the time "17:20" to a python datetime.time object."""
    timestr = timestring.strip()
    hh = int(timestr[0:2])
    mm = int(timestr[3:5])
    return time(hour=hh, minute=mm)


def corr_environlen_by_timelen(timelist, environmentlist):
    """Checks and corrects time-list versus environment-list. It always adjusts
    environment data to time: truncates too long environment-lists and
    expands too short environment lists according to last entry.
    DOESN'T CHECK CONTENT."""
    envl = copy.deepcopy(environmentlist)
    tl = len(timelist)
    evlast = envl[-1]
    while tl > len(envl):
        envl.append(evlast)
    while tl < len(envl):
        del(envl[-1])
    return envl


def readinputdatalines(inpfilename):
    """ Reads inputdata from inputfilename and returns result as
    dictionary"""
    allowed_keywords = ALLOWED_INPUTFILE_KEYWORDS
    try:
        csvfile = open(inpfilename, 'r')
    except:
        print("ERROR: No existing inputfile " + inpfilename)
        print("Exit")
        sys.exit(aerr.ERRORS['INPFILEERROR'])
    reader = csv.reader(csvfile, delimiter=';')
    data = []  # empty at start
    for row in reader:
        if len(row) > 0:  # ingore empty lines
            keyword = row[0].strip().upper()
            if keyword[0] != '#':  # ignore comment lines
                linedat = row[1:]
                dataline = (keyword, linedat)
                data.append(dataline)
    # check data and add to an tuple for return
    returndict = {}
    for d in data:
        keyw = d[0]
        try:
            if keyw in allowed_keywords:
                if keyw == 'STATION':
                    station = d[1][0].strip()
                    returndict[keyw] = station.upper()
                if keyw == 'DATE1':
                    datum1 = str2date(d[1][0].strip())
                    returndict[keyw] = datum1
                if keyw == 'DATE2':
                    datum2 = str2date(d[1][0].strip())
                    returndict[keyw] = datum2
                if keyw == 'DATESTEP':
                    datestep = int(d[1][0].strip())
                    returndict[keyw] = datestep  # int(datestep)
                if keyw == 'TIMES':
                    times = []
                    for t in d[1]:
                        tx = str2hhmm(t)
                        times.append(tx)
                    returndict[keyw] = times
                if keyw == 'CLOUDS':
                    clouds = []
                    for c in d[1]:
                        cx = int(c)
                        clouds.append(cx)
                    returndict[keyw] = clouds
        except:
            print("ERROR: Wrong inputdata in " + inpfilename)
            print("at record {} :".format(d))
            print("Exitcode {}".format(
                aerr.ERRORS['INPUTFILE_NOVALIDDATALINE']))
            sys.exit(aerr.ERRORS['INPUTFILE_NOVALIDDATALINE'])
    # print returndict
    return returndict


def process_inputdict(inputdict):
    """Processes input dictionary after reading the inputfile.csv and
    does some checks and corrections and returns the necessary data fields.
    Works for programs that need envirnoment data (cloudiness, etc..)"""
    station = inputdict['STATION']
    station = station.upper()
    pydate = inputdict['DATE1']
    times = inputdict['TIMES']
    cl = corr_environlen_by_timelen(times, inputdict['CLOUDS'])
    clouds = []
    for c in cl:  # limit cloudiness betwenn 0 and 10(!) octus
        c = min(10, max(c, 0))
        clouds.append(c)
    clouds = np.array(clouds)
    environment = (clouds,)
    t1 = times[0]  # first time in list
    t2 = times[-1]  # last time in list
    if t2 >= t1:
        print("ERROR: " + str(t2) + " >= " + str(t1))
        print ("Time span 24 hours or more not allowed!")
        print("Check inputfile. - Exit!")
        sys.exit(aerr.ERRORS['TIMESPANERROR'])
    pydatetime_array = []
    pystartdate = pydate  # old day
    pydate2 = pydate + timedelta(days=1)  # new day
    for t in times:
        pydate = pystartdate
        if t < t1:  # change of date
            pydate = pydate2
        ptd = adat.combinepydatepytime(pydate, t)
        pydatetime_array.append(ptd)
    pystartdatetime = pydatetime_array[0]
    pyenddatetime = pydatetime_array[-1]
    pytimestep = None
    return (station, pystartdate, pystartdatetime, pyenddatetime,
            pytimestep, pydatetime_array, environment)


def readstationlist(locindicator, statfilename=STATFILE):
    """Searches the ICAO location indicator in statfile and returns
    station data as dictionary in the format:
    {'station':stat, 'name':name, 'lat':lat, 'lon':lon, 'elev':elev}

    In case of error it aborts with an appropriate message to stdout"""
    try:
        with open(statfilename, 'rb') as f:
            reader = csv.reader(f)
            found = False
            for row in reader:
                if row[0] == locindicator:
                    found = True
                    stat = row[0]
                    name = row[1]
                    lat = row[9]
                    lon = row[10]
                    elev = row[8]
                    return {'station': stat, 'name': name, 'lat': lat,
                            'lon': lon, 'elev': elev}
            if not found:
                    print("Error: Location Indicator not found: " +
                          locindicator)
                    sys.exit(aerr.ERRORS['NOLOCINDICATOR'])
    except IOError:
        print("Error: no stationfile: " + statfilename)
        sys.exit(aerr.ERRORS['NOSTATFILE'])


def readstationlistdb(locationindicator, db=STATIONDATABASE):
    """Reads from an appropriate sqlite3-Database *db* the location
         indicator *locindicator* and its geographical data and
         returns a dictionary with the form {'station': string,
         'name': string, 'lat': float, 'lon': float, 'elev': float}
    """
    locind = locationindicator
    if not op.isfile(db):
        print("ERROR: Stationlist-DB " + db + " not available")
        sys.exit(aerr.ERRORS['NOSTATDATABASE'])
    con = sq.connect(db)
    c = con.cursor()
    try:
        c.execute('SELECT * FROM stationlist WHERE indicator=?',
                  (locind,))
    except:
        print("ERROR: " + db +
              " is no datatbase or not the correct station database")
        sys.exit(102)
    l = c.fetchone()
    con.close()
    if l is None:
        print("ERROR: Station " + locind + " is not in datatbase " + db)
        sys.exit(aerr.ERRORS['NOLOCINDICATOR'])
    d = {'station': l[0], 'name': l[1],
         'lat': (l[2] + l[3] / 60.0 + l[4] / 3600.0),
         'lon': (l[5] + l[6] / 60.0 + l[7] / 3600.0),
         'elev': l[8]}
    return d


def ephemdate2ddhhmm_str(d):
    """pyephem rise/set-date to yymmdd string"""
    if str(d) == "-- --":
        dstr = "   --:--"
    elif str(d) == '++ ++':
        dstr = "   ++:++"
    else:
        try:
            psr = d.tuple()
            dstr = "{:0>2d} {:0>2d}:{:0>2.0f}".format(psr[2], psr[3],
                                                      psr[4] + float(psr[5]) / 60.)
        except:
            # some quere event
            dstr = "   xx:xx"
    return dstr


def ephemdate2ddhhmm_csv(d):
    try:
        psr = d.tuple()
        dstr1 = "{:0>4d}-{:0>2d}-{:0>2d}".format(psr[0], psr[1], psr[2])
        dstr2 = "{:0>2d}:{:0>2.0f}".format(
            psr[3],
            round(psr[4] + float(psr[5] / 60.0)))
        dstr = dstr1 + "; " + dstr2
    except:
        # some quere event
        dstr = "9999-12-31; 00:00"
    return dstr


def ephemdate2yymmdd_str(d):
    """pypehem next/previous date to string"""
    if str(d) == "-- --":
        dstr = "--.--.--"
    elif str(d) == '++ ++':
        dstr = "++.++.++"
    else:
        try:
            psr = d.tuple()
            dstr = "{0:0>2d}.{1:0>2d}.{2:0>2d}".format(psr[2], psr[1],
                                                       psr[0] % 100)
        except:
            # some quere event
            dstr = "xx.xx.xx"
    return dstr


def ephemdate2yymmdd_csv(d):
    try:
        psr = d.tuple()
        dstr = "{0:0>4d}-{1:0>2d}-{2:0>2d}".format(psr[0], psr[1],
                                                   psr[2])
    except:
        # some quere event
        dstr = "9999-12-31"
    return dstr


def makepydatetimelist(pystart, pyend, pystep):
    """ Returns a list of python datetime-objects that
    starts with pystart, increments with pystep and does
    end with as last element of the list less-equal pyend."""
    pydatetime = []
    pydt = pystart
    while pydt <= pyend:
        pydatetime.append(pydt)
        pydt = pydt + pystep
    return pydatetime


def prevnext_moonevent(ephem_date, moonriseset_tuple, csv=False):
    if csv:
        RISECHAR = 'r;'
        SETCHAR = 's;'
    else:
        RISECHAR = 'r '
        SETCHAR = 's '
    mrst = moonriseset_tuple
    try:
        if mrst['prevrise'] > mrst['prevset']:
            # prev. rise after prev. set
            rs0 = [RISECHAR, mrst['prevrise']]
        else:
            rs0 = [SETCHAR, mrst['prevset']]
    except:
        rs0 = None, None
    try:
        if mrst['nextset'] < mrst['nextrise']:
            rs1 = [SETCHAR, mrst['nextset']]
        else:
            rs1 = [RISECHAR, mrst['nextrise']]
    except:
        rs1 = None, None
    return rs0, rs1


def astro_dayline(
    location_indicator=LOCINDICATORDEFAULT, actualdate=date.today(),
        stationdatabase=STATIONDATABASE, printout=True):
    """ Calculates a line of data with usefull informations concerning
    sunrise, moonrise etc.. for one night.

    Arguments:

    * location indicator
    * date in python style
    * stationdatabase (usually not to be passed)
    * printout ... boolean, default True (for historical compatibility
      reasons)

    Returns a single formatted string with sunset, ECET, ENET, BNMT,
    BCMT, sunrise, moonrise or set (whatever is earlier), moonphase,
    next new moon, next full moon if printout == False
    """

    datetimeutc = adat.Date(actualdate)
    location = readstationlistdb(location_indicator)
    # lat = al.deg2rad(float(location['lat']))
    # lon = al.deg2rad(float(location['lon']))
    locnoon, locmidnight, srs, ctw, ntw, atw, mrs, mphase, mfn, nfm, nnm = al.calc_astro_night(
        location, datetimeutc)
    m0, m1 = prevnext_moonevent(locmidnight, mrs)
    ds0 = ephemdate2ddhhmm_str(srs['prevsunset'])
    ds1 = ephemdate2ddhhmm_str(srs['nextsunrise'])
    dc0 = ephemdate2ddhhmm_str(ctw['prevsunset'])
    dc1 = ephemdate2ddhhmm_str(ctw['nextsunrise'])
    dn0 = ephemdate2ddhhmm_str(ntw['prevsunset'])
    dn1 = ephemdate2ddhhmm_str(ntw['nextsunrise'])
    dm0 = ephemdate2ddhhmm_str(m0[1])
    dm1 = ephemdate2ddhhmm_str(m1[1])
    dm0 = m0[0] + dm0
    dm1 = m1[0] + dm1
    dmf = ephemdate2yymmdd_str(nfm)
    dmn = ephemdate2yymmdd_str(nnm)
    datal1 = "{}  {}  {}  {}  {}  {} ".format(ds0, dc0, dn0, dn1, dc1, ds1)
    datal2 = " {}  {}  {:3.0f}%  {}   {}".format(dm0, dm1, mphase, dmf, dmn)
    dataline = " {}  {}  {}  {}  {}  {} | {}  {}  {:3.0f}%  {}   {}".format(
        ds0, dc0, dn0, dn1, dc1, ds1, dm0, dm1, mphase, dmf, dmn)
    if printout:
        print actualdate.strftime("%Y-%m-%d | ") + dataline
    else:
        return datal1, datal2


def astro_dayline_csv(
    location_indicator=LOCINDICATORDEFAULT, actualdate=date.today(),
        stationdatabase=STATIONDATABASE):
    """Calculates a line of csv formatted data with usefull informations concerning
    sunrise, moonrise etc.. for one night.

    Arguments:

    * location indicator
    * date in python style
    * stationdatabase (usually not to be passed)

    Returns:

    A single csv formatted string (semicolon separated) with sunset, ECET,
    ENET, BNMT, BCMT, sunrise, moonrise or set (whatever is earlier),
    moonphase, next new moon, next full moon.

    * locind ... location indicator
    * date ... Day of interrest
    * ssdate  ... Sunset Date
    * sstime ... Sunset Time
    * ecetdate ... ECET
    * ecettime
    * enetdate ... ENET
    * enettime
    * bmntdate ... BMNT
    * bmnttime
    * bcmtdate ... BCMT
    * bcmttime
    * srdate ... Sunrise
    * srtime
    * srflag1 ... moon event 1: rise r, set s
    * moon1date
    * moon1time
    * srflag2 ... moon event 2
    * moon2date
    * moon2time
    * moonphase ... moon phase (%) at mean midnight
    * nextfullm ... next full moon
    * nextnewm ... next new moon
    """

    datetimeutc = adat.Date(actualdate)
    location = readstationlistdb(location_indicator)
    locnoon, locmidnight, srs, ctw, ntw, atw, mrs, mphase, mfn, nfm, nnm = al.calc_astro_night(
        location, datetimeutc)
    m0, m1 = prevnext_moonevent(locmidnight, mrs, csv=True)
    ds0 = ephemdate2ddhhmm_csv(srs['prevsunset'])
    ds1 = ephemdate2ddhhmm_csv(srs['nextsunrise'])
    dc0 = ephemdate2ddhhmm_csv(ctw['prevsunset'])
    dc1 = ephemdate2ddhhmm_csv(ctw['nextsunrise'])
    dn0 = ephemdate2ddhhmm_csv(ntw['prevsunset'])
    dn1 = ephemdate2ddhhmm_csv(ntw['nextsunrise'])
    dm0 = ephemdate2ddhhmm_csv(m0[1])
    dm1 = ephemdate2ddhhmm_csv(m1[1])
    dm0 = m0[0] + dm0
    dm1 = m1[0] + dm1
    dmf = ephemdate2yymmdd_csv(nfm)
    dmn = ephemdate2yymmdd_csv(nnm)

    dataline = "{}; {}; {}; {}; {}; {}; {}; {}; {:3.0f}; {}; {}".format(
        ds0, dc0, dn0, dn1, dc1, ds1, dm0, dm1, mphase, dmf, dmn)
    print(
        location_indicator +
        "; " +
        actualdate.strftime(
            "%Y-%m-%d") +
        "; " +
        dataline)


def maketime_line(dates, trailer="Time     :  ",
                  formatstring="  {:02d}:{:02d} "):
    """ Makes a line of times with hours and minutes for output"""
    line = trailer
    for d in dates:
        d = adat.Date(d)
        d = d.tuple()
        mm = int(round(d[4] + d[5] / 60.))
        hh = d[3]
        # to mitigate the crazy problems of 17:59:59.9999999999999993
        # there is no better way I can see :-(
        if mm >= 60:
            hh = hh + 1
            mm = mm - 60
        str2 = formatstring.format(hh, mm)
        line = line + str2
    return line


def makegeneraldata_line(data, trailer="Temperature: ",
                         formatstring="  {:5.1f} ", maxout=999.9):
    """General data line string with trailer and single data as in
    formatstring"""
    line = trailer
    for d in data:
        if d > maxout:
            d = maxout
        str1 = formatstring.format(d)
        line = line + str1
    return line


def moon_daylines(moondata):
    """Returnes datalines for moon position and relative illumination"""
    # D o   c a l c u l a t i o n s
    dates, altitude, azimuth, size, magnitude, phase = moondata
    # illumination normal to the moon "beam" calculated of appar.magnitude
    norm_illum = al.mag2lux(magnitude) * 1000.  # lx -> mlx
    # illumination projected to a plane and extincted by atmosphere
    illum = al.moon_illuminance(altitude, norm_illum)  # allready mlx!
    # relative illumination normed with the maximum illumination by
    # a mean zenithal moon
    rel_illum = illum / (1000 * al.ILLUM_FULLMOON_MEAN)  # lx -> mlx
    # P r e p a r e   d a t a   l i n e s
    header = maketime_line(dates)
    alt_deg = al.rad2deg(altitude)
    altstr = makegeneraldata_line(alt_deg, "Alt      : ", "    {:4.0f}")
    azi_deg = al.rad2deg(azimuth)
    azstr = makegeneraldata_line(azi_deg, "Azi      : ", "    {:4.0f}")
    relillumstr = makegeneraldata_line(
        100. * rel_illum,
        "rel.ill.%: ",
        "   {:5.0f}")
    return header, altstr, azstr, relillumstr


def moon_daylines_csv(moondata):
    """Returnes datalines for moon position and relative illumination as csv"""
    # D o   c a l c u l a t i o n s
    dates, altitude, azimuth, size, magnitude, phase = moondata
    # illumination normal to the moon "beam" calculated of appar.magnitude
    norm_illum = al.mag2lux(magnitude) * 1000.  # lx -> mlx
    # illumination projected to a plane and extincted by atmosphere
    illum = al.moon_illuminance(altitude, norm_illum)  # allready mlx!
    # relative illumination normed with the maximum illumination by
    # a mean zenithal moon
    rel_illum = illum / (1000 * al.ILLUM_FULLMOON_MEAN)  # lx -> mlx
    dates, altitude, azimuth, size, magnitude, phase = moondata
    alt_deg = al.rad2deg(altitude)
    az_deg = al.rad2deg(azimuth)
    # D o   C S V - O u t p u t
    header = maketime_line(dates, 'time; ', ' {:02d}:{:02d};')
    altstr = makegeneraldata_line(alt_deg, "altitude; ", "{:5.1f}; ")
    azstr = makegeneraldata_line(az_deg, "azimuth; ", "{:5.1f}; ")
    relillumstr = makegeneraldata_line(
        rel_illum * 100.,
        "relillum; ",
        "{:5.1f}; ")
    return header, altstr, azstr, relillumstr


def illum_daylines(moondata, sundata, environment):
    """Returnes datalines for illuminances"""
    # D o   c a l c u l a t i o n s
    clouds = environment[0]
    dates = moondata[0]
    m_illum, s_illum, twl_illum, brutto_illum, netto_illuminance =\
        al.calc_illumination(moondata, sundata, environment)
    # p r e p a r e   o u t p u t   l i n e s
    header = maketime_line(dates)
    cloudline = makegeneraldata_line(clouds, 'Clouds   :  ', '     {:2.0f} ')
    # millumline = makegeneraldata_line(m_illum, 'M.Ill.   :  ', '  {:5.1f} ')
    # sillumline = makegeneraldata_line(s_illum, 'S.Ill.   :  ', '  {:5.1f} ')
    # twlillumline = makegeneraldata_line(
    #    twl_illum,
    #    'Tw.Ill.  :  ',
    #    '  {:5.1f} ')
    nettoillumline = makegeneraldata_line(
        netto_illuminance,
        'Ill. mlx :  ',
        '  {:5.1f} ')
    return header, cloudline, nettoillumline


def illum_daylines_csv(moondata, sundata, environment):
    """Returnes csv-datalines for illuminances"""
    # D o   c a l c u l a t i o n s
    clouds = environment[0]
    dates = moondata[0]
    m_illum, s_illum, twl_illum, brutto_illum, netto_illuminance =\
        al.calc_illumination(moondata, sundata, environment)
    # p r e p a r e   o u t p u t   l i n e s
    header = maketime_line(dates, 'time ; ', ' {:02d}:{:02d};')
    cloudline = makegeneraldata_line(clouds, 'clouds; ', ' {:2.0f}; ')
    nettoillumline = makegeneraldata_line(
        netto_illuminance,
        'illuminance;  ',
        '{:5.1f}; ')
    return header, cloudline, nettoillumline


def stationdata_lines(locationindicator=LOCINDICATORDEFAULT):
    """ Returns string lines with station information"""
    statdat = readstationlistdb(locationindicator)
    line1 = u"Station: {}  {}".format(statdat['station'], statdat['name'])
    line2 = u"Latitude: {:8.4f}  Longitude: {:7.4f} degree   Elevation: {:4.0f} m".format(
        statdat['lat'], statdat['lon'], statdat['elev'])
    return line1, line2


def stationdata_lines_csv(locationindicator=LOCINDICATORDEFAULT):
    """Returns csv-string lines with station information """
    statdat = readstationlistdb(locationindicator)
    line1 = u"station; {};  name; {};".format(
        statdat['station'], statdat['name'])
    line2 = u"latitude; {:8.4f}; longitude; {:7.4f}; elevation; {:4.0f};"\
        .format(statdat['lat'], statdat['lon'], statdat['elev'])
    return line1, line2


if __name__ == "__main__":
    aconst.version()
