#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Creation date: 2014-05-17

"""**Admont Good Night Vision library** for standard astronomical
calculations based on `pyephem <http://rhodesmill.org/pyephem/>`_.
Also some routines for night time illuminance calculations are
provided.

It can read geographical data from a textfile or a sqlite3-database
that is part of the distribution. It comprises sun rise and set,
twilights, moon rise, sun position, moon position and moon phases.

| |copy| 2015, D. Thaler
| All rights reserved
| www: http://agnitevision.foehnwall.at/modules.html
| E-mail: info at foehnwall dot at
"""

import sys
from datetime import date, datetime, timedelta, time
try:
    import numpy as np
except:
    ERROR_np = "ERROR: no lib numpy available. Install numpy e.g. from: "
    ERROR_np = ERROR_np + "http://scipy.org/"
    print(ERROR_np)
    sys.exit(-1)
try:
    import ephem
except:
    ERROR_ep = "ERROR: no lib ephem available. Install ephem e.g. from: "
    ERROR_ep = ERROR_ep + "https://pypi.python.org/pypi/pyephem/"
    print(ERROR_ep)
    sys.exit(-1)

import adnitevision_errlib as aerr
import adnitevision_datlib as adat

__version__ = aerr.VERSION
__copyright__ = "(c) D.Thaler 2015"
__license__ = "EUPL - https://joinup.ec.europa.eu/software/page/eupl"

VERSION = __version__  # : Program version

# SUN and TWILIGHTS (fits USNO-Data)
SUNRISESET = '-0:50'  # :sun radius 0°16' + standard atmospheric refraction 0°34'
TWLCIVIL = ' -6:00'  # :sun height civil twilight -6°
TWLNAUTIC = '-12:00'  # :sun height nautical twilight -12°
TWLASTRO = '-18:00'  # :sun height asronomical twilight -18°

HCIVIL = -6.0  # sun altitude for civil
HNAUTIC = -12.0  # sun altitude for nautical
HASTRONOMIC = -18.0  # sun altitude for austronomical

BACKGROUNDILLUMINATION = 1.5  #: mlx, rough estimate clear sky
SUNSETILLUMINATION = 5.5e5  #: mlx , rough estimate sunset clear sky

# Maybe for some later use
# MOONDIAMETER = 3476.0*1000 #:diamater of moon in m
# MOONRADIUS = MOONDIAMETER/2. #:radius of moon in m
# AU = 149597870700.0
# ASTRONOMICAL_UNIT = AU #: astronomical unit in  m

DEG2RAD = np.pi / 180.0

# apparent magnitude of the full moon
# http://en.wikipedia.org/wiki/Apparent_magnitude#Example:_Sun_and_Moon
MAG_FULLMOON_MAX = - \
    12.92  #: maximum magnitude of full moon ( http://en.wikipedia.org/wiki/Apparent_magnitude )
MAG_FULLMOON_MEAN = - \
    12.74  #: mean magnitude of full moon ( http://en.wikipedia.org/wiki/Apparent_magnitude )


def mag2lux(m):
    """Converts astronomical magnitude to illuminace (Beleuchtungsstaerke)
    in lux normal to the direction of the object according to the
    formula given at
    http://members.ziggo.nl/jhm.vangastel/Astronomy/Formules.pdf

    """
    return 10.0 ** ((-14.18 - m) / 2.5)

ILLUM_FULLMOON_MAX = mag2lux(MAG_FULLMOON_MAX)
                             #: maximum brightness of full moon in zenith [lx]
ILLUM_FULLMOON_MEAN = mag2lux(MAG_FULLMOON_MEAN)
                              #: mean brightness of full moon in zenith [lx]
Q_TRANSPARENCYDICT = {
    'very clear': 0.95,
    'clear': 0.90,
    'medium': 0.85,
    'turbid': 0.80,
    'very turbid': 0.75}
Q_TRANSPARENCY = Q_TRANSPARENCYDICT[
    'medium']  #: transparency of the atmosphere (1 ... full transparency, 0 .. full opacity)

#
# Q_TRANSPARENCY = 1.0
#

MINIMUM_CLOUDFACTOR = 0.333  #: reduction of relative illuminance at overcast sky (N=8)


def deg2rad(x):
    """Degrees to Radians"""
    return np.deg2rad(x)


def rad2deg(x):
    """Radians to Degrees"""
    return np.rad2deg(x)


def relatmass(altrad):
    """ Relative atmospheric mass according to Pickering(2002) as
    a function of altitude in rad."""
    altdeg = rad2deg(altrad)
    idx = altdeg < 0.0
    altdeg[idx] = 0.0
    q = altdeg + 244.0 / (165.0 + 57.0 * np.power(altdeg, 1.1))
    M = 1.0 / np.sin(deg2rad(q))
    return M


def degrees(angle):
    """Passing thru the ephem.degrees function. Creates a degree object"""
    return ephem.degrees(angle)


def localnoon(lon='0.0', timeutc=datetime.utcnow()):
    """Calculates local mean noon in UTC"""
    long = rad2deg(float(lon))
    dd = ephem.Date(round(timeutc))
    dt = float(long) / 15.0 / 24.
    ddnew = ephem.Date(dd - dt)
    return ddnew


def sunriseset(location, ctime, horizon='-0:50'):
    """Calculates sun rise and set for a *location* and around the
    *time* above the *horizon*. Time should be the mean local noon
    or at least a time where the sun is above the horizon. It Returns
    a dictionary with {'prevsunset':set_prev, 'prevsunrise':rise_prev,
    'nextsunset':set_next,'nextsunrise':rise_next}
    """
    NVRUP = '-- --'
    NVRDOWN = '++ ++'
    loc = ephem.Observer()
    loc.horizon = horizon
    # print "sunriseset-horizon", horizon
    loc.pressure = 0.0  # refraction is accounted for in horizon
    loc.lon = str(location['lon'])
    loc.lat = str(location['lat'])
    loc.elevation = float(location['elev'])
    loc.date = ctime
    sun = ephem.Sun()
    try:
        set_prev = loc.previous_setting(sun, use_center=True)
        set_next = loc.next_setting(sun, use_center=True)
        rise_prev = loc.previous_rising(sun, use_center=True)
        rise_next = loc.next_rising(sun, use_center=True)
    except ephem.NeverUpError:
        set_prev, rise_prev, set_next, rise_next = NVRUP, NVRUP, NVRUP, NVRUP
    except ephem.AlwaysUpError:
        set_prev, rise_prev, set_next, rise_next = NVRDOWN, NVRDOWN, NVRDOWN, NVRDOWN
    sunsrdict = {'prevsunset': set_prev, 'prevsunrise': rise_prev,
                 'nextsunset': set_next, 'nextsunrise': rise_next}
    return sunsrdict


def moonriseset(location, ctime, horizon='-0:34'):
    """Calculates moon rise and set for a *location* and around the
    time *ctime* above the *horizon*. Standard horizon accounts for
    atmospheric refraction.

    **For reasonable results the time must be chosen
    carefully.** If the moon has e.g. allready set, the next set will be
    tommorow though he previous rise has been today. So not all events
    of the day and the night might be caught.

    The function returns a dictionary with the results.
    The results can be derived by:

    * dict['prevrise]
    * dict['prevset']
    * dict['nextrise']
    * dict['nextset']
    """
    NVRUP = '-- --'
    NVRDOWN = '++ ++'
    loc = ephem.Observer()
    loc.horizon = horizon
    loc.pressure = 0.0  # refraction is accounted for with the value of the horizon
    loc.lon = str(location['lon'])
    loc.lat = str(location['lat'])
    loc.elevation = float(location['elev'])
    loc.date = ctime
    moon = ephem.Moon()
    # print "M: ctime ", ctime
    # print "M: loc.date ",loc.date
    try:
        set_prev = loc.previous_setting(moon)
        # print "M: set_prev",set_prev
        set_next = loc.next_setting(moon)
        # print "M: set_next",set_next
        rise_prev = loc.previous_rising(moon)
        # print "M: rise_prev",rise_prev
        rise_next = loc.next_rising(moon)
        # print "M: rise_next",rise_next
    except ephem.NeverUpError:
        set_prev, rise_prev, set_next, rise_next = NVRUP, NVRUP, NVRUP, NVRUP
    except ephem.AlwaysUpError:
        set_prev, rise_prev, set_next, rise_next = NVRDOWN, NVRDOWN, NVRDOWN, NVRDOWN
    moonsrdict = {'prevset': set_prev, 'prevrise': rise_prev,
                  'nextset': set_next, 'nextrise': rise_next}
    return moonsrdict


def moonphase(ctime):
    """ Phase of the moon in percent of the apparent area of the full moon"""
    m = ephem.Moon()
    m.compute(ctime)
    return m.phase


def moonfullnew(ctime):
    """Calculates previous and next main moon phases (new, 1.quater, full,
    last quater). Returnes a dictionary {'prev_new_moon':pnm,
    'next_new_moon':nnm, 'prev_1st_quarter':pfq,
    'next_1st_quarter':nfq, 'prev_full_moon':pfm,
    'next_full_moon':nfm, 'prev_last_quarter':plq,
    'next_last_quarter':nlq}

    """
    pnm = ephem.previous_new_moon(ctime)
    nnm = ephem.next_new_moon(ctime)
    pfq = ephem.previous_first_quarter_moon(ctime)
    nfq = ephem.next_first_quarter_moon(ctime)
    pfm = ephem.previous_full_moon(ctime)
    nfm = ephem.next_full_moon(ctime)
    plq = ephem.previous_last_quarter_moon(ctime)
    nlq = ephem.next_last_quarter_moon(ctime)
    mphasedict = {'prev_new_moon': pnm, 'next_new_moon': nnm,
                  'prev_1st_quarter': pfq, 'next_1st_quarter': nfq,
                  'prev_full_moon': pfm, 'next_full_moon': nfm,
                  'prev_last_quarter': plq, 'next_last_quarter': nlq}
    return mphasedict


def sundata(location, pydatetimelist):
    """Calculates sun altitude, azimuth, size and magnitude for a given
    location between to dates with a given time increment.

    *Parameters*:

    * location ... dictionary with the geographic data of the location
      as returned by e.g. stationlistdb
    * pydatetimelist ... a list with standard python datetime-objects

    *Returns*: tuple of (dates, altitude, azimuth,size,magnitude)

    * dates ... numpy array with pyephem.Dates-Data
    * altitude ... numpy array with altitudes in pyephem.degrees -
      data
    * azimuth ... numpy array with azimuth in pyephem.degrees - data
    * size ...  numpy array with azimuth in pyephem.size - data
    * magnitude ... numpy array with azimuth in pyephem.mag - data

    """
    # convert python datetimes to pyephem datetimes
    dates = []
    for pdt in pydatetimelist:
        edt = ephem.Date(pdt)
        dates.append(edt)
    dates = np.array(dates)
    #
    altitude = np.empty_like(dates)
    azimuth = np.empty_like(dates)
    size = np.empty_like(dates)
    magnitude = np.empty_like(dates)
    ort = ephem.Observer()
    ort.lat = str(location['lat'])
    ort.lon = str(location['lon'])
    ort.elevation = float(location['elev'])
    ort.pressure = 0.0  # *NECESSARY* to avoid strange refration effects
    sun = ephem.Sun(ort)
    ind = 0
    for d in dates:
        ort.date = d
        sun.compute(ort)
        altitude[ind] = sun.alt
        azimuth[ind] = sun.az
        size[ind] = sun.size
        magnitude[ind] = sun.mag
        ind = ind + 1
    return dates, altitude, azimuth, size, magnitude


def moondata(location, pydatetimelist):
    """Calculates moon altitude, azimuth, size and magnitude for a given
    location between to dates with a given time increment.

    *Parameters*:

    * location ... dictionary with the geographic data of the location
      as returned by e.g. stationlistdb
    * pydatetimelist ... a list with standard python datetime-objects

    *Returns*: tuple of (dates, altitude, azimuth, size, magnitude)

    * dates ... numpy array with pyephem.Dates-Data
    * altitude ... numpy array with altitudes in pyephem.degrees -
      data (radiant)!
    * azimuth ... numpy array with azimuth in pyephem.degrees - data
      (radiant!)
    * size ...  numpy array with azimuth in pyephem.size - data
    * magnitude ... numpy array with azimuth in pyephem.mag - data

    """
    # convert python datetimes to pyephem datetimes
    dates = []
    for pdt in pydatetimelist:
        edt = ephem.Date(pdt)
        dates.append(edt)
    dates = np.array(dates)
    #
    altitude = np.empty_like(dates)
    azimuth = np.empty_like(dates)
    size = np.empty_like(dates)
    magnitude = np.empty_like(dates)
    phase = np.empty_like(dates)
    ort = ephem.Observer()
    ort.lat = str(location['lat'])
    ort.lon = str(location['lon'])
    ort.elevation = float(location['elev'])
    ort.pressure = 0.0  # *NECESSARY* to avoid strange refraction effects
    moon = ephem.Moon(ort)
    ind = 0
    for d in dates:
        ort.date = d
        moon.compute(ort)
        altitude[ind] = moon.alt
        azimuth[ind] = moon.az
        size[ind] = moon.size
        magnitude[ind] = moon.mag
        phase[ind] = moon.phase
        ind = ind + 1
    return dates, altitude, azimuth, size, magnitude, phase


def moonsizemag(location, date):
    ort = ephem.Observer()
    ort.lat = str(location['lat'])
    ort.lon = str(location['lon'])
    ort.elevation = float(location['elev'])
    ort.pressure = 0.0  # *NECESSARY* to avoid strange refraction effects
    moon = ephem.Moon(ort)
    ort.date = date
    moon.compute(ort)
    return moon.size, moon.mag


def clearskyilluminance(altrad, illum0):
    """Calculates the illuminance on a horizontal surface  with

    *Input-Parameters*

    * altrad ... altitude of the celestial body in rad(!)

    * illum0 ... illuminance appropriate units of the celestial body
      normal to the beam witout atmospheric attenuation

    *Outputs*

    * illuminance ... at a horizontal surface in whatever units illum0 has
      been passed (with no extra atmospheric attenuation)
    """
    # altrad = deg2rad(altdeg)
    illum_noatm_proj = illum0 * np.sin(altrad)
    return illum_noatm_proj


def globalilluminance(altrad, illum0):
    # Moonilluminace + Backgroundilluminantion#
    ill1 = clearskyilluminance(altrad, illum0)
    idx = altrad < 0.0
    ill1[idx] = 0.0
    M = relatmass(altrad)
    ill2 = ill1 * np.power(Q_TRANSPARENCY, M)
    illum = BACKGROUNDILLUMINATION + ill2
    return illum


def moon_illuminance(altrad, illum0):
    """ Calculates the illuminace of the moon dependent
    on moon altitude and with atmospheric extinction."""
    ill1 = clearskyilluminance(altrad, illum0)
    idx = altrad < 0.0
    ill1[idx] = 0.0
    M = relatmass(altrad)
    return ill1 * np.power(Q_TRANSPARENCY, M)


def sun_illuminance(altrad, illum0):
    """ Calculates the illuminace of the sun dependent
    on moon altitude and with atmospheric extinction."""
    ill1 = clearskyilluminance(altrad, illum0)
    idx = altrad < 0.0
    ill1[idx] = 0.0
    M = relatmass(altrad)
    return ill1 * np.power(Q_TRANSPARENCY, M)


def twilight_illuminance(sunaltitude_rad):
    """Estimates the clear sky illuminance for twilights.
    Starts with global constant SUNSETILLUMINACE and ends with
    ZERO at astronomical wilight. A modified exponetial approach is used.
    **ATTENTION**: Twilightilluminace containes BACKGROUNDILLUMINATION and
    so background illuminance does not need any further consideration!"""
    alpha = -np.log(SUNSETILLUMINATION / BACKGROUNDILLUMINATION) / HASTRONOMIC
    h = rad2deg(sunaltitude_rad)  # convert to degrees
    tw_il = SUNSETILLUMINATION * np.exp(alpha * h)  # BACKGROUNDILLUMINATION
    idx = (h < HASTRONOMIC)
    tw_il[idx] = BACKGROUNDILLUMINATION
    idx = (h > 0.0)
    tw_il[idx] = SUNSETILLUMINATION
    return tw_il


def linear_cloudfactor(N):
    """Reduction of illuminance of an astronomical source by cloud cover
    from N=0 to N=8 octusses. Linear approach, cf(N=0) = 1,
    cf(N=8) = MINIMUM_CLOUDFACTOR"""
    cf = 1. - (1 - MINIMUM_CLOUDFACTOR) / 8.0 * N
    idx1 = (cf < 0)
    cf[idx1] = 0.0
    idx2 = (cf > 1)
    cf[idx2] = 1.0
    return cf


def exponential_cloudfactor(N):
    """Reduction of illuminance of an astronomical source by cloud cover
    from N=0 to N=8 octusses. Exponential approach, cf(N=0) = 1,
    cf(N=8) = MINIMUM_CLOUDFACTOR"""
    A = -np.log(MINIMUM_CLOUDFACTOR) / 8.0
    cf = np.exp(-A * N)
    idx2 = (cf > 1)
    cf[idx2] = 1.0
    return cf


def linexp_cloudfactor(N):
    """Reduction of illuminance of an astronomical source by cloud cover
    from N=0 to N=8 octusses. Mixed approach, cf(N=0) = 1,
    cf(N=8) = MINIMUM_CLOUDFACTOR"""
    cf1 = linear_cloudfactor(N)
    cf2 = exponential_cloudfactor(N)
    return 0.5 * (cf1 + cf2)


def cloudfactor(N):
    """Reduction of illuminance of an astronomical source by cloud cover
    from N=0 to N=8 octusses."""
    return linear_cloudfactor(N)
    # return exponential_cloudfactor(N)
    # return linexp_cloudfactor(N)


def calc_astro_night(location, datetimeutc):
    """Takes

    * location dictionary with the geographic data of the location as
      returned by e.g. stationlistdb
    * pyephem style utc time

    and returns a tuple with all usefull astronomic standard data
    structures or the midnight following the noon of the day:

    * local noon
    * local midnight 12 hors after noon
    * sun rise and set
    * civil twilight
    * nautical twilight
    * astronomical twilight
    * moon rise and set
    * moon phase in percent
    * moon full and new
    * next full moon
    * next new moon
    """
    lon = deg2rad(float(location['lon']))
    locnoon = localnoon(lon, datetimeutc)
    locmidnight = adat.Date(locnoon + 12. / 24.)
    srs = sunriseset(location, locmidnight, SUNRISESET)
    ctw = sunriseset(location, locmidnight, TWLCIVIL)
    ntw = sunriseset(location, locmidnight, TWLNAUTIC)
    atw = sunriseset(location, locmidnight, TWLASTRO)
    mrs = moonriseset(location, locmidnight)
    mphase = moonphase(locmidnight)
    mfn = moonfullnew(locmidnight)
    nfm = mfn['next_full_moon']
    nnm = mfn['next_new_moon']
    return locnoon, locmidnight, srs, ctw, ntw, atw, mrs, mphase, mfn, nfm, nnm


def calc_illumination(moondata, sundata, environment):
    """ Calculates night illumination dependent on moondata, sundata
    and environment data."""
    clouds = environment[0]
    mdates, maltitude, mazimuth, msize, mmagnitude, mphase = moondata
    sdates, saltitude, sazimuth, ssize, smagnitude = sundata
    # illumination normal to the moon "beam" calculated of appar.magnitude
    m_norm_illum = mag2lux(mmagnitude) * 1000.  # lx -> mlx
    # illumination projected to a plane and extincted by atmosphere
    m_illum = moon_illuminance(maltitude, m_norm_illum)  # allready mlx!
    # illumination normal to the sun beam calculated of appar.magnitude
    s_norm_illum = mag2lux(smagnitude) * 1000.  # lx -> mlx
    # illumination projected to a plane and extincted by atmosphere
    s_illum = sun_illuminance(saltitude, s_norm_illum)  # allready mlx!
    # twilight illuminance
    twl_illum = twilight_illuminance(saltitude)
    # all together without clouds (background in twilight included!)
    brutto_illum = m_illum + twl_illum
    environment_factor = cloudfactor(clouds)
    # all together with clouds
    netto_illum = environment_factor * brutto_illum
    return m_illum, s_illum, twl_illum, brutto_illum, netto_illum


def calc_illumination_from_scratch(location, pydatetimelist):
    pass


def version():
    """ Prints version and copyright message on standard output"""
    print("Module: " + sys.argv[0])
    print("Version: " + __version__)
    print("License: " + __license__ + " " + __copyright__)

if __name__ == "__main__":
    version()
